package lightcontainer.domains.client;

import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.interfaces.StoreProcessorFFE;
import lightcontainer.protocol.ProtocolReader;
import lightcontainer.protocol.ProtocolWriter;

import java.io.*;
import java.net.Socket;
import java.nio.charset.StandardCharsets;
import java.util.Objects;

/**
 * StoreProcessor
 *
 * <!> MULTICAST CLIENT <!>
 * Class communicating with the storebackend and
 * processing the tasks in the FileFrontEnd
 *
 * @version     1.1
 * @since       1.0
 *
 * @see         Runnable
 * @see         AutoCloseable
 * @author      Jérémi NIHART <j.nihart@student.helmo.be>
 */
public class StoreProcessor extends Thread implements AutoCloseable {
    // Variables
    private final StoreProcessorFFE fileFrontEnd;
    private final Socket store;
    private final String domain;
    private boolean client_run;

    private BufferedReader reader;
    private PrintWriter writer;
    private String currentCommand;
    private ProtocolRepository protocolRep;

    // Constructor
    public StoreProcessor(Socket socket, String domain, StoreProcessorFFE ffe, ProtocolRepository protocolRep) {
        this.domain = domain;
        this.fileFrontEnd = ffe;
        this.store = socket;
        this.protocolRep = protocolRep;
        this.client_run = false;
        initStore();
    }

    /**
     * Initialise the Store's Reader and Writer.
     *
     * @see BufferedReader
     * @see PrintWriter
     * @since 1.0
     */
    private void initStore() {
        try {
            this.reader = new BufferedReader(new InputStreamReader(
                    this.store.getInputStream(),
                    StandardCharsets.UTF_8
            ));
            this.writer = new PrintWriter(new OutputStreamWriter(
                    this.store.getOutputStream(),
                    StandardCharsets.UTF_8
            ), true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Thread Function
     * Start the dialogue with the storebackend.
     *
     * @since 1.0
     */
    @Override
    public void run() {
        this.client_run = true;
        while (this.client_run) {
            try {
                alertAvalaible();

                // Request
                ProtocolWriter.ProtocolResult requestResult = protocolRep.executeWriter(this.currentCommand);
                this.writer.write(requestResult.getCommand());
                requestResult.write(this.store.getOutputStream());


                // Response
                String responseCommand = this.reader.readLine();
                if (responseCommand != null)
                System.out.println("StoreBackEnd: " + responseCommand);
                ProtocolReader.ProtocolResult responseResult = protocolRep.executeReader(responseCommand);
                System.out.println("StoreBackEnd response to client: " + responseResult.getResultCommand());


            } catch (IOException ignore) { }
        }
    }

    /**
     * Permet de demander au StoreBackEnd d'effectuer une commande
     * @param command La commande à effectuer
     */
    public void executeCommand(String command) {
        synchronized (this) {
            this.currentCommand = command;
            this.notify();
        }
    }

    /**
     * Permet de déclarer le StoreBackEnd disponible
     */
    private void alertAvalaible() {
        synchronized (this) {
            this.currentCommand = null;

            fileFrontEnd.onStoreAvailable(this);

            try {
                this.wait();
            } catch (InterruptedException e) { e.printStackTrace(); }
        }
    }

    /**
     * AutoClosable Function
     * Close the Storage thread and resources.
     *
     * @since 1.0
     */
    @Override
    public void close() {
        if (this.client_run) {
            try {
                this.client_run = false;
                this.store.close();
            } catch (IOException ignored) { }
        }
    }

    /**
     * Compare two StoreProcessor object to check if equals.
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        StoreProcessor that = (StoreProcessor) o;
        return Objects.equals(domain, that.domain);
    }

    /**
     * Get hash code
     */
    @Override
    public int hashCode() {
        return Objects.hash(domain);
    }
}
